/* wrap.m     Simple custom wrappers for Slide24 example
 *
 * Copyright (c) 2011 Psellos   http://psellos.com
 * Licensed under the MIT license:
 *     http://www.opensource.org/licenses/mit-license.php
 */
# import <stdlib.h>
# import <objc/runtime.h>
# import <Quartzcore/Quartzcore.h>
# import <UIKit/UIKit.h>
# import <caml/mlvalues.h>
# import <caml/memory.h>
# import <caml/alloc.h>
# import <caml/callback.h>
# import "wrap.h"

static value Val_fourtuple(double a, double b, double c, double d);
static void Fourtuple_val(double *ap, double *bp, double *cp, double *dp,
                        value tuple);
static unsigned Bitset_val(value bitset);
static value Val_UIColor(UIColor *color);
static UIColor *UIColor_val(value colorval);
static value Val_CGColor(CGColorRef color);
static CGColorRef CGColor_val(value colorval);
static SEL SEL_val(value selectval);


/* Represent raw ObjC object pointers in OCaml as nativeint values.
 *
 * Note: to keep things simple, we retain ObjC values and don't release
 * them.  It works for this example because the few ObjC objects
 * accessible from the OCaml side are never destroyed.
 * 
 * In a more dynamic setting, need to gain control when the OCaml
 * objects are GCed and release the corresponding ObjC objects.
 */
static value wrapRawObjC(id obj) {
    CAMLparam0();
    [obj retain];
    CAMLreturn(caml_copy_nativeint((intnat) obj));
}

static id unwrapRawObjC(value objcval) {
    return (id) Nativeint_val(objcval);
}

/* Need to be able to determine the OCaml object that wraps a given ObjC
 * object.  In the simple setup here, the association is permanent and
 * there are very few objects involved.  So we can use a little table.
 * (Getting the ObjC object that wraps an OCaml object is easier; we
 * just use the container method of Wrappee.t.)
 */
static struct {
    id objc_obj;
    value ocaml_obj;
} g_ocaml_wrapper[64];
static int g_ocwrap_ct = 0;


static value find_ocaml_wrapper(id obj) {
    int i;

    for(i = 0; i < g_ocwrap_ct; i++)
        if(g_ocaml_wrapper[i].objc_obj == obj)
            return g_ocaml_wrapper[i].ocaml_obj;
    return 0;
}


static value wrapObjC(char *wrapcs, id obj) {
    /* Create a new OCaml object that wraps the given ObjC object.
     * wrapcs gives the name of the closure for creating an OCaml
     * wrapper of the correct type.  A closure is registered by each
     * OCaml class that wraps an ObjC class (i.e., each subclass of
     * Wrapper.t).
     */
    CAMLparam0();
    CAMLlocal1(val);
    value *closure;

    closure = (value *) caml_named_value(wrapcs);
    if(closure == NULL) {
        /* Can't create an instance of an unregistered class.  If this
         * happens, things will go very bad from here.
         */
        fprintf(stderr, "wrapObjC lookup failure for %s\n", wrapcs);
        fflush(stderr);
        CAMLreturn(Val_unit); /* Not type safe, should never happen */
    }
    val = caml_callback(*closure, wrapRawObjC(obj));

    /* Enter the wrapping into the table.
     */
    g_ocaml_wrapper[g_ocwrap_ct].objc_obj = obj;
    g_ocaml_wrapper[g_ocwrap_ct].ocaml_obj = val;
    caml_register_global_root(&g_ocaml_wrapper[g_ocwrap_ct].ocaml_obj);
    g_ocwrap_ct++;

    CAMLreturn(val);
}

static id unwrapObjC(value obj) {
    /* Return the ObjC object wrapped by the given OCaml object.
     */
    CAMLparam1(obj);
    CAMLlocal1(objcval);

    objcval =
        caml_callback(
            caml_get_public_method(obj, caml_hash_variant("contents")),
            obj);
    CAMLreturnT(id, unwrapRawObjC(objcval));
}


/* Subclasses of WrapOCaml wrap OCaml objects as ObjC objects.
 */
@implementation WrapOCaml {
}

@synthesize contents;

- (WrapOCaml *) init
{
    /* Create a new OCaml object that is wrapped by self.  We match the
     * OCaml class to our own class by name--simple but effective.  We
     * then call the closure for creating a wrapped object, which is
     * registered by each OCaml class that can be wrapped an ObjC class
     * (i.e., each subclass of Wrappee.t).
     */
    const char *classname;
    char closurename[128];
    value *closure;

    if((self = [super init]) != nil) {
        classname = object_getClassName(self);
        if(strlen(classname) + 8 >= sizeof closurename) {
            fprintf(stderr, "[WrapOCaml init]: class name too long: %s\n",
                    classname);
            fflush(stderr);
            return nil;
        }
        strcpy(closurename, classname);
        strcat(closurename, ".wrapped");
        closure = (value *) caml_named_value(closurename);
        if(closure == NULL) {
            fprintf(stderr, "[WrapOCaml init]: lookup failure for %s\n",
                    closurename);
            fflush(stderr);
            return nil;
        }
        contents = caml_callback(*closure, wrapRawObjC(self));
        caml_register_global_root(&contents);
    }
    return self;
}

- (void) dealloc
{
    if(contents != 0) {
        caml_remove_global_root(&contents);
        contents = 0;
    }
    [super dealloc];
}

@end

value Wrappee_performSelector_withDelay_(value objcval, value selval,
                    value delayval)
/* nativeint -> string -> float -> unit */
{
    /* Any subclass of Wrappee.t can queue a selector to be delivered on
     * the main thread at a later time.  If the delay is 0.0, it's a way
     * of yielding to the other parties in the thread (i.e., the GUI)
     * before continuing.
     */
    CAMLparam3(objcval, selval, delayval);

    NSObject *wrapper = (NSObject *) unwrapRawObjC(objcval);
    [wrapper performSelector: SEL_val(selval)
            withObject: nil
            afterDelay: Double_val(delayval)];

    CAMLreturn(Val_unit);
}


/* OCaml objects accessed from ObjC.
 */

@implementation Slide24ctlr : WrapOCaml
{
}

@dynamic window;

- (UIWindow *) window
{
    CAMLparam0();
    CAMLlocal2(selfval, windowval);

    selfval = [self contents];
    windowval =
        caml_callback(
            caml_get_public_method(selfval, caml_hash_variant("window")),
            selfval);
    CAMLreturnT(UIWindow *, unwrapObjC(windowval));
}

- (void) setWindow: (UIWindow *) aWindow
{
    CAMLparam0();
    CAMLlocal2(selfval, windowval);

    selfval = [self contents];
    windowval = wrapObjC("UiWindow.wrap", (id) aWindow);
    (void)
        caml_callback2(
            caml_get_public_method(selfval, caml_hash_variant("setWindow'")),
            selfval,
            windowval);
    CAMLreturn0;
}

@dynamic shuffle;

- (UIButton *) shuffle
{
    CAMLparam0();
    CAMLlocal2(selfval, buttonval);

    selfval = [self contents];
    buttonval =
        caml_callback(
            caml_get_public_method(selfval, caml_hash_variant("shuffle")),
            selfval);
    CAMLreturnT(UIButton *, unwrapObjC(buttonval));
}

- (void) setShuffle: (UIButton *) aButton
{
    CAMLparam0();
    CAMLlocal2(selfval, buttonval);

    selfval = [self contents];
    buttonval = wrapObjC("UiButton.wrap", (id) aButton);
    (void)
        caml_callback2(
            caml_get_public_method(selfval, caml_hash_variant("setShuffle'")),
            selfval,
            buttonval);
    CAMLreturn0;
}

@dynamic solve;

- (UIButton *) solve
{
    CAMLparam0();
    CAMLlocal2(selfval, buttonval);

    selfval = [self contents];
    buttonval =
        caml_callback(
            caml_get_public_method(selfval, caml_hash_variant("solve")),
            selfval);
    CAMLreturnT(UIButton *, unwrapObjC(buttonval));
}

- (void) setSolve: (UIButton *) aButton
{
    CAMLparam0();
    CAMLlocal2(selfval, buttonval);

    selfval = [self contents];
    buttonval = wrapObjC("UiButton.wrap", (id) aButton);
    (void)
        caml_callback2(
            caml_get_public_method(selfval, caml_hash_variant("setSolve'")),
            selfval,
            buttonval);
    CAMLreturn0;
}

- (void) applicationDidFinishLaunching: (UIApplication *) anApplication
{
    CAMLparam0();
    CAMLlocal2(selfval, applval);

    selfval = [self contents];
    if((applval = find_ocaml_wrapper(anApplication)) == 0)
        applval = wrapObjC("UiApplication.wrap", (id) anApplication);
    (void)
        caml_callback2(
            caml_get_public_method(
                selfval,
                caml_hash_variant("applicationDidFinishLaunching'")),
            selfval,
            applval);
    CAMLreturn0;
}

- (void) applicationWillResignActive: (UIApplication *) anApplication
{
    CAMLparam0();
    CAMLlocal2(selfval, applval);

    selfval = [self contents];
    if((applval = find_ocaml_wrapper(anApplication)) == 0)
        applval = wrapObjC("UiApplication.wrap", (id) anApplication);
    (void)
        caml_callback2(
            caml_get_public_method(
                selfval,
                caml_hash_variant("applicationWillResignActive'")),
            selfval,
            applval);
    CAMLreturn0;
}

- (void) applicationDidBecomeActive: (UIApplication *) anApplication
{
    CAMLparam0();
    CAMLlocal2(selfval, applval);

    selfval = [self contents];
    if((applval = find_ocaml_wrapper(anApplication)) == 0)
        applval = wrapObjC("UiApplication.wrap", (id) anApplication);
    (void)
        caml_callback2(
            caml_get_public_method(
                selfval,
                caml_hash_variant("applicationDidBecomeActive'")),
            selfval,
            applval);
    CAMLreturn0;
}

- (void) touchTile: (UIButton *) aButton
{
    CAMLparam0();
    CAMLlocal2(selfval, buttonval);

    selfval = [self contents];
    if((buttonval = find_ocaml_wrapper(aButton)) == 0)
        /* Can't touch unknown button.
         */
        return;
    (void)
        caml_callback2(
            caml_get_public_method(selfval, caml_hash_variant("touchTile'")),
            selfval,
            buttonval);
    CAMLreturn0;
}

- (IBAction) shuffle: (UIButton *) button
{
    CAMLparam0();
    CAMLlocal2(selfval, buttonval);

    selfval = [self contents];
    if((buttonval = find_ocaml_wrapper(button)) == 0)
        /* Can't touch unknown button.
         */
        return;
    (void)
        caml_callback2(
            caml_get_public_method(selfval, caml_hash_variant("shuffle'")),
            selfval,
            buttonval);
    CAMLreturn0;
}

- (IBAction) solve: (UIButton *) button
{
    CAMLparam0();
    CAMLlocal2(selfval, buttonval);

    selfval = [self contents];
    if((buttonval = find_ocaml_wrapper(button)) == 0)
        /* Can't touch unknown button.
         */
        return;
    (void)
        caml_callback2(
            caml_get_public_method(selfval, caml_hash_variant("solve'")),
            selfval,
            buttonval);
    CAMLreturn0;
}

- (void) runsolve
{
    CAMLparam0();
    CAMLlocal1(selfval);

    selfval = [self contents];
    (void)
        caml_callback(
            caml_get_public_method(selfval, caml_hash_variant("runsolve")),
            selfval);

    CAMLreturn0;
}

- (void) timerTick: (NSTimer *) aTimer
{
    CAMLparam0();
    CAMLlocal2(selfval, timerval);

    selfval = [self contents];
    if((timerval = find_ocaml_wrapper(aTimer)) == 0)
        /* Can't report a timer tick if the timer isn't known.
         */
        return;
    (void)
        caml_callback2(
            caml_get_public_method(selfval, caml_hash_variant("timerTick'")),
            selfval,
            timerval);
    CAMLreturn0;
}

@end


/* ObjC objects accessed from OCaml.
 */

/* NSTimer
 */

value NSTimer_scheduledTimer5(value timeInterval, value target,
                    value selectval, value userInfo, value repeats)
    /* For simplicity, the target must be wrapped (by an ObjC object)
     * already.  It's OK if userInfo isn't wrapped; we treat this as
     * nil.
     */
{
    CAMLparam5(timeInterval, target, selectval, userInfo, repeats);
    CAMLlocal2(targro, uiro);

    double ti = Double_val(timeInterval);
    targro = 
        caml_callback(
            caml_get_public_method(target, caml_hash_variant("container")),
            target);
    id targ = unwrapRawObjC(targro);
    if(targ == nil) {
        /* If the target wasn't already wrapped, things will go very bad
         * from here.
         */
        fprintf(stderr, "NSTimer_scheduledTimer5: unwrapped target\n");
        fflush(stderr);
        CAMLreturn(Val_unit); /* Not type safe, should never happen */
    }
    SEL sel = SEL_val(selectval);
    uiro =
        caml_callback(
            caml_get_public_method(userInfo, caml_hash_variant("container")),
            userInfo);
    id ui = unwrapRawObjC(uiro);
    int rep = Bool_val(repeats);

    NSTimer *timer =
        [NSTimer scheduledTimerWithTimeInterval: ti
                 target: targ
                 selector: sel
                 userInfo: ui
                 repeats: rep ];
    CAMLreturn(wrapObjC("NsTimer.wrap", (id) timer));
}


/* CALayer
 */

value CALayer_masksToBounds(value objcval)
/* nativeint -> bool */
{
    CAMLparam1(objcval);

    CALayer *layer = unwrapRawObjC(objcval);

    CAMLreturn(Val_bool([layer masksToBounds]));
}

value CALayer_setMasksToBounds_(value objcval, value boolval)
/* nativeint -> bool -> unit */
{
    CAMLparam2(objcval, boolval);

    CALayer *layer = unwrapRawObjC(objcval);
    [layer setMasksToBounds: Int_val(boolval)];

    CAMLreturn(Val_unit);
}

value CALayer_cornerRadius(value objcval)
/* nativeint -> float */
{
    CAMLparam1(objcval);

    CALayer *layer = unwrapRawObjC(objcval);

    CAMLreturn(caml_copy_double([layer cornerRadius]));
}

value CALayer_setCornerRadius_(value objcval, value floatval)
/* nativeint -> float -> unit */
{
    CAMLparam2(objcval, floatval);

    CALayer *layer = unwrapRawObjC(objcval);
    [layer setCornerRadius: Double_val(floatval)];

    CAMLreturn(Val_unit);
}

value CALayer_borderWidth(value objcval)
/* nativeint -> float */
{
    CAMLparam1(objcval);

    CALayer *layer = unwrapRawObjC(objcval);

    CAMLreturn(caml_copy_double([layer borderWidth]));
}

value CALayer_setBorderWidth_(value objcval, value floatval)
/* nativeint -> float -> unit */
{
    CAMLparam2(objcval, floatval);

    CALayer *layer = unwrapRawObjC(objcval);
    [layer setBorderWidth: Double_val(floatval)];

    CAMLreturn(Val_unit);
}

value CALayer_borderColor(value objcval)
/* nativeint -> color */
{
    CAMLparam1(objcval);

    CALayer *layer = unwrapRawObjC(objcval);

    CAMLreturn(Val_CGColor([layer borderColor]));
}

value CALayer_setBorderColor_(value objcval, value colorval)
/* nativeint -> color -> unit */
{
    CAMLparam2(objcval, colorval);

    CALayer *layer = unwrapRawObjC(objcval);
    CGColorRef cgcolor = CGColor_val(colorval);
    [layer setBorderColor: cgcolor];

    CAMLreturn(Val_unit);
}

/* UIView
 */

value UIView_beginAnimations_context_(value animidval, value ctxval)
/* string -> int -> unit */
{
    CAMLparam2(animidval, ctxval);

    NSString *animid = [NSString stringWithUTF8String: String_val(animidval)];
    [UIView beginAnimations: animid context: (void *) Int_val(ctxval)];

    CAMLreturn(Val_unit);
}

value UIView_commitAnimations(value unitval)
/* unit -> unit */
{
    CAMLparam1(unitval);
    [UIView commitAnimations];
    CAMLreturn(Val_unit);
}

value UIView_setAnimationDuration_(value duraval)
/* float -> unit */
{
    CAMLparam1(duraval);
    [UIView setAnimationDuration: Double_val(duraval)];
    CAMLreturn(Val_unit);
}


/* UIButton
 */

value UIButton_buttonWithType_(value btyval)
/* buttonType -> UiButton.t */
{
    CAMLparam1(btyval);

    UIButton *button = [UIButton buttonWithType: Int_val(btyval)];

    CAMLreturn(wrapObjC("UiButton.wrap", (id) button));
}

value UIButton_layer(value objcval)
/* nativeint -> CALayer.t */
{
    CAMLparam1(objcval);
    CAMLlocal1(layerval);
    CALayer *layer;

    UIButton *button = unwrapRawObjC(objcval);
    layer = [button layer];

    if((layerval = find_ocaml_wrapper(layer)) == 0)
        layerval = wrapObjC("CaLayer.wrap", (id) layer);

    CAMLreturn(layerval);
}

value UIButton_frame(value objcval)
/* nativeint -> rect */
{
    CAMLparam1(objcval);
    CAMLlocal1(res);

    UIButton *button = unwrapRawObjC(objcval);
    CGRect frame = [button frame];
    res =
        Val_fourtuple(frame.origin.x, frame.origin.y,
                        frame.size.width, frame.size.height);
    CAMLreturn(res);
}

value UIButton_addTarget_action_forControlEvents_(value objcval, value targval,
                    value actval, value evval)
/* nativeint -> #Wrappee.t -> string -> controlEvents -> unit */
{
    CAMLparam4(objcval, targval, actval, evval);
    CAMLlocal1(targro);
    char *prime;

    UIButton *button = unwrapRawObjC(objcval);
    targro =
        caml_callback(
            caml_get_public_method(targval, caml_hash_variant("container")),
            targval);
    id targ = unwrapRawObjC(targro);
    [button addTarget: targ
            action: SEL_val(actval)
            forControlEvents: Bitset_val(evval)];

    CAMLreturn(Val_unit);
}

value UIButton_enabled(value objcval)
/* nativeint -> bool */
{
    CAMLparam1(objcval);

    UIButton *button = unwrapRawObjC(objcval);

    CAMLreturn(Val_bool([button isEnabled]));
}

value UIButton_setEnabled_(value objcval, value boolval)
/* nativeint -> bool -> unit */
{
    CAMLparam2(objcval, boolval);

    UIButton *button = unwrapRawObjC(objcval);
    [button setEnabled: Bool_val(boolval)];

    CAMLreturn(Val_unit);
}

value UIButton_setFrame_(value objcval, value rectval)
/* nativeint -> rect -> unit */
{
    CAMLparam2(objcval, rectval);
    CGRect frame;

    UIButton *button = unwrapRawObjC(objcval);
    frame.origin.x = Double_val(Field(rectval, 0));
    frame.origin.y = Double_val(Field(rectval, 1));
    frame.size.width = Double_val(Field(rectval, 2));
    frame.size.height = Double_val(Field(rectval, 3));

    [button setFrame: frame];
    CAMLreturn(Val_unit);
}

value UIButton_setTitle_forState_(value objcval, value titleval, value stval)
/* nativeint -> string -> controlStates -> unit */
{
    CAMLparam3(objcval, titleval, stval);

    UIButton *button = unwrapRawObjC(objcval);
    NSString *title = [NSString stringWithUTF8String: String_val(titleval)];

    [button setTitle: title forState: Bitset_val(stval)];
    CAMLreturn(Val_unit);
}

value UIButton_setTitleColor_forState_(value objcval, value colorval,
                        value stval)
/* nativeint -> color -> controlStates -> unit */
{
    CAMLparam3(objcval, colorval, stval);

    UIButton *button = unwrapRawObjC(objcval);
    UIColor *color = UIColor_val(colorval);

    [button setTitleColor: color forState: Bitset_val(stval)];
    CAMLreturn(Val_unit);
}

value UIButton_titleColorForState_(value objcval, value stval)
/* nativeint -> controlStates -> color */
{
    CAMLparam2(objcval, stval);

    UIButton *button = unwrapRawObjC(objcval);
    UIColor *color = [button titleColorForState: Bitset_val(stval)];

    CAMLreturn(Val_UIColor(color));
}

value UIButton_titleForState_(value objcval, value stval)
/* nativeint -> controlStates -> string */
{
    CAMLparam2(objcval, stval);

    UIButton *button = unwrapRawObjC(objcval);
    NSString *title = [button titleForState: Bitset_val(stval)];

    CAMLreturn(caml_copy_string([title UTF8String]));
}

value UIButton_titleSize(value objcval)
/* nativeint -> float */
{
    /* This is a synthetic method.  In a richer setting you might
     * provide access to the title label [UIButton titleLabel] and wrap
     * the UILabel and UIFont classes.
     */
    CAMLparam1(objcval);

    UIButton *button = unwrapRawObjC(objcval);
    CGFloat size = [[[button titleLabel] font] pointSize];

    CAMLreturn(caml_copy_double((double) size));
}

value UIButton_setTitleSize_(value objcval, value ptsval)
/* nativeint -> float -> unit */
{
    /* Another synthetic method like the previous.
     */
    CAMLparam2(objcval, ptsval);

    UIButton *button = unwrapRawObjC(objcval);
    UIFont *font = [UIFont systemFontOfSize: Double_val(ptsval)];
    [[button titleLabel] setFont: font];

    CAMLreturn(Val_unit);
}

value UIButton_setBackgroundImage_forState_(value objcval, value imageval,
                    value stval)
/* nativeint -> image -> controlStates -> unit */
{
    CAMLparam3(objcval, imageval, stval);

    UIButton *button = unwrapRawObjC(objcval);
    NSString *name = [NSString stringWithUTF8String: String_val(imageval)];
    UIImage *image = [UIImage imageNamed: name];
    [button setBackgroundImage: image forState: Bitset_val(stval)];

    CAMLreturn(Val_unit);
}

value UIButton_currentTitle(value objcval)
/* nativeint -> string */
{
    CAMLparam1(objcval);

    UIButton *button = unwrapRawObjC(objcval);
    NSString *title = [button currentTitle];

    CAMLreturn(caml_copy_string([title UTF8String]));
}

value UIButton_currentTitleColor(value objcval)
/* nativeint -> color */
{
    CAMLparam1(objcval);

    UIButton *button = unwrapRawObjC(objcval);
    UIColor *color = [button currentTitleColor];

    CAMLreturn(Val_UIColor(color));
}


/* UIWindow
 */

value UIWindow_addSubview_(value objcval, value subvval)
/* nativeint -> nativeint -> unit */
{
    /* For simplicity, the OCaml type of subvval is UiButton.t.  In a
     * more complex setting it could be any subclass of UIView.
     */
    CAMLparam2(objcval, subvval);

    UIWindow *window = unwrapRawObjC(objcval);
    UIView *subview = unwrapRawObjC(subvval);
    [window addSubview: subview];

    CAMLreturn(Val_unit);
}


/* Utilities
 */
static value Val_fourtuple(double a, double b, double c, double d)
{
    CAMLparam0();
    CAMLlocal1(res);

    res = caml_alloc_tuple(4);
    Store_field(res, 0, caml_copy_double(a));
    Store_field(res, 1, caml_copy_double(b));
    Store_field(res, 2, caml_copy_double(c));
    Store_field(res, 3, caml_copy_double(d));

    CAMLreturn(res);
}


static void Fourtuple_val(double *ap, double *bp, double *cp, double *dp,
                        value tuple)
{
    CAMLparam1(tuple);

    *ap = Double_val(Field(tuple, 0));
    *bp = Double_val(Field(tuple, 1));
    *cp = Double_val(Field(tuple, 2));
    *dp = Double_val(Field(tuple, 3));

    CAMLreturn0;
}


static unsigned Bitset_val(value bitsetval)
{
    CAMLparam1(bitsetval);
    unsigned res;

    res = 0;
    while(bitsetval != Val_emptylist) {
        res |= 1 << Int_val(Field(bitsetval, 0));   /* car */
        bitsetval = Field(bitsetval, 1);            /* cdr */
    }

    CAMLreturnT(unsigned, res);
}


static value Val_UIColor(UIColor *color)
{
    return Val_CGColor([color CGColor]);
}


static UIColor *UIColor_val(value colorval)
{
    CAMLparam1(colorval);
    double a, b, c, d;
    UIColor *res;

    Fourtuple_val(&a, &b, &c, &d, colorval);
    res = [UIColor colorWithRed: a green: b blue: c alpha: d];

    CAMLreturnT(UIColor *, res);
}


static value Val_CGColor(CGColorRef cgcolor)
{
    CAMLparam0();
    double rgba[4];
    int ct, i;

    /* Colors will always be RGBA, but it doesn't hurt to be careful.
     */
    for(i = 0; i < 3; i++)
        rgba[i] = 0.0;
    rgba[3] = 1.0;

    if((ct = CGColorGetNumberOfComponents(cgcolor)) > 4)
        ct = 4;

    const CGFloat *comps = CGColorGetComponents(cgcolor);
    for(i = 0; i < ct; i++)
        rgba[i] = comps[i];

    CAMLreturn(Val_fourtuple(rgba[0], rgba[1], rgba[2], rgba[3]));
}


static CGColorRef CGColor_val(value colorval)
{
    CAMLparam1(colorval);

    /* (Defer colorspace handling to UIColor.)
     */
    UIColor *color = UIColor_val(colorval);

    CAMLreturnT(CGColorRef, [color CGColor]);
}


static SEL SEL_val(value selectval)
{
    /* Translate the OCaml string into an ObjC selector.
     *
     * Conventionally we use '\'' in OCaml where ':' appears in ObjC.
     * So here we translate back.
     */
    char *prime;
    SEL res;

    char *ocamlsel = String_val(selectval);
    char *buf = malloc(strlen(ocamlsel) + 1);
    strcpy(buf, ocamlsel);
    prime = buf;
    while((prime = strchr(prime, '\'')) != NULL)
        *prime = ':';
    res = sel_registerName(buf);
    free(buf);
    return res;
}
